/*

    KMatrix: Simulates moving matrix like in famous film.
    Copyright (C) 2021  Vitaliy Krylov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <sys/ksys.h>

#define bool unsigned char
#define true 1
#define false 0
#define nullptr 0
#define dword unsigned int

#define CONSOLE_WIDTH 80
#define CONSOLE_HEIGHT 25

#define SYSFN04_FONT_W 8
#define SYSFN04_FONT_H 14

#define TH_STACK_SIZE 1024

void mymemcpy(void* dest, void* src, unsigned int bytes) {
	unsigned int i = 0;
	for (; i < bytes / sizeof(dword); ++i) {
		((dword*)dest)[i] = ((dword*)src)[i];
	}
	for (unsigned int j = i * sizeof(dword); j < bytes; ++j) {
		dest[j] = src[j];
	}
}

// Slightly edited implementation from glibc
unsigned int rand_r_seed = 0;
int rand_r() {
	unsigned int *seed = &rand_r_seed;
	unsigned int next = *seed;
	int result;

	next *= 1103515245;
	next += 12345;
	result = (unsigned int) (next / 65536) % 2048;

	next *= 1103515245;
	next += 12345;
	result <<= 10;
	result ^= (unsigned int) (next / 65536) % 1024;

	next *= 1103515245;
	next += 12345;
	result <<= 10;
	result ^= (unsigned int) (next / 65536) % 1024;

	*seed = next;

	return result;
}

int create_thread(void* func) {
	void* th_stack = malloc(TH_STACK_SIZE); // Allocate memory for thread stack
	if (!th_stack) {
		_ksys_debug_puts("[KMatrix] Thread stack allocation failed.\n");
		return 1;
	}
	unsigned tid = _ksys_create_thread(func, th_stack+TH_STACK_SIZE); // Create new thread with entry "main"
	if (tid == -1) {
		_ksys_debug_puts("[KMatrix] Unable to create a new thread.\n");
		free(th_stack);
		return 2;
	}
	return 0;
}

//<Queue>
typedef struct {
	unsigned int queue_size;
	unsigned int sizeoftype;
	void* data;
	unsigned int data_maxsize;
	unsigned int data_begin;
	unsigned int data_size;
} Queue;

//queue_size = 0 if want dynamic size choosing
Queue* Q_alloc(unsigned int sizeoftype, unsigned int queue_size) {
	Queue* this = (Queue*)malloc(sizeof(Queue));
	if (!this) {
		_ksys_debug_puts("[KMatrix] Q_alloc(): Queue allocation failed.\n");
		return nullptr;
	}
	this->sizeoftype = sizeoftype;
	this->data_begin = 0;
	this->data_size = 0;
	if (queue_size) this->queue_size = queue_size;
	else this->queue_size = 0x8000 / sizeoftype;
	this->data_maxsize = this->queue_size;
	this->data = (void*)malloc(this->data_maxsize * sizeoftype);
	if (!this->data) {
		_ksys_debug_puts("[KMatrix] Q_alloc(): Queue.data allocation failed.\n");
		free(this);
		return nullptr;
	}
	return this;
}

void Q_free(Queue* this) {
	free(this->data);
	free(this);
}

//$value -- it is a reference to what you want to add
void Q_push(Queue* this, const void* value) {
	if (this->data_size == this->data_maxsize) {
		void* olddata = this->data;
		this->data = (void*)malloc((this->data_maxsize + this->queue_size) * this->sizeoftype);
		if (!this->data) {
			_ksys_debug_puts("[KMatrix] Q_push(): Data allocation failed.\n");
			this->data = olddata;
			return;
		}

		int data_i = 0;
		for (int i = this->data_begin; i < this->data_maxsize; ++i) mymemcpy(this->data + this->sizeoftype * (data_i++), olddata + this->sizeoftype * i, this->sizeoftype);
		for (int i = 0; i < this->data_begin; ++i) mymemcpy(this->data + this->sizeoftype * (data_i++), olddata + this->sizeoftype * i, this->sizeoftype);

		free(olddata);
		this->data_maxsize += this->queue_size;
		this->data_begin = 0;
	}
	if (this->data_begin + this->data_size < this->data_maxsize) mymemcpy(this->data + this->sizeoftype * (this->data_begin + this->data_size), value, this->sizeoftype);
	else mymemcpy(this->data + this->sizeoftype * (this->data_begin + this->data_size - this->data_maxsize), value, this->sizeoftype);
	++this->data_size;
}

void Q_pop(Queue* this, void* returnValue) {
	if (this->data_size > 0) {
		if (returnValue) mymemcpy(returnValue, this->data + this->sizeoftype * this->data_begin, this->sizeoftype);
		++this->data_begin;
		if (this->data_begin == this->data_maxsize)
			this->data_begin = 0;
		--this->data_size;
	} else _ksys_debug_puts("[KMatrix] Q_pop(): Wrong use.\n");
}

//WARNING: Be sure not to pass nullptr
void Q_front(Queue* this, void* returnValue) {
	if (this->data_size > 0) mymemcpy(returnValue, this->data + this->sizeoftype * this->data_begin, this->sizeoftype);
	else _ksys_debug_puts("[KMatrix] Q_front(): Wrong use.\n");
}

unsigned int Q_size(Queue* this) {
	return this->data_size;
}

//NOTICE: Doesn't free used memory. Will use it for next pushes.
void Q_clear(Queue* this) {
	this->data_begin = 0;
	this->data_size = 0;
}

//WARNING: Be sure not to pass nullptr
void Q_at(Queue* this, int i, void* returnValue) {
	if (this->data_begin + i < this->data_maxsize)
		return mymemcpy(returnValue, this->data + this->sizeoftype * (this->data_begin + i), this->sizeoftype);
	return mymemcpy(returnValue, this->data + this->sizeoftype * (this->data_begin + i - this->data_maxsize), this->sizeoftype);
}
//</Queue>

int reportConIOError() {
	_ksys_debug_puts("[KMatrix] Unable to initialize console.\n");
	return 1;
}

typedef struct {
	int pos;
	int length;
	char* sequence; // if first element is '\0' then its an empty row
} MatrixRow;

#include <time.h>
#include <string.h>

int main(int argc, char** argv) {
	int fgcolor = 32, bgcolor = 40, hlcolor = 37, matrixDelay = 10;
	int fgcolorrgb = 0, bgcolorrgb = 0, hlcolorrgb = 0;
	bool fgcolor_rgboverriden = false, bgcolor_rgboverriden = false, hlcolor_rgboverriden = false;
	bool runInShell = false;
	for (int i = 1; i < argc; ++i) {
		if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
			if (con_init() && reportConIOError()) return 1; // Init fail
			(*con_set_title)("KMatrix -- Matrix for KolibriOS");
			printf("Usage: %s [OPTION]...\n", argv[0]);
			puts("Simulates moving matrix like in famous film.\n\n");

			puts("\t-F, --foreground\t<COLOR>\tset color of letters;\n");
			puts("\t\t\tavailable colors: black, red, green, brown,\n");
			puts("\t\t\tblue, purple, turqoise, white\n");
			puts("\t\t\tor (available only in graphic mode):\n");
			puts("\t\t\t<R,G,B>   For example: '-F 255,0,255'\n");
			puts("\t-B, --background\t<COLOR>\tset color of background;\n");
			puts("\t\t\tavailable colors: black, red, green, brown,\n");
			puts("\t\t\tblue, purple, turqoise, white\n");
			puts("\t\t\tor (available only in graphic mode):\n");
			puts("\t\t\t<R,G,B>   For example: '-B 255,0,255'\n");
			puts("\t-H, --highlight\t<COLOR>\tset color of lead symbols;\n");
			puts("\t\t\tavailable colors: black, red, green, brown,\n");
			puts("\t\t\tblue, purple, turqoise, white\n");
			puts("\t\t\tor (available only in graphic mode):\n");
			puts("\t\t\t<R,G,B>   For example: '-H 255,0,255'\n");
			puts("\t-S, --speed\t<VALUE>\tset speed of falling letters;\n");
			puts("\t\t\tit must be an integer and be more than 0;\n");
			puts("\t\t\tdetermines how many falls will be per second;\n");
			puts("\t\t\tnotice: formula of waiting is floor(100 / speed),\n");
			puts("\t\t\tit means that some speed values will give same looking\n");
			puts("\t--shell\t\trun this program in shell\n");
			puts("\t--help\t\tdisplay this help and exit\n");
			puts("\t--version\t\toutput version information and exit\n\n");

			puts("Defaults when nothing has overriden:\n");
			puts("\t--foreground green\n");
			puts("\t--background black\n");
			puts("\t--highlight white\n");
			puts("\t--speed 10\n");
			(*con_exit)(0);
			return 0;
		} else if (!strcmp(argv[i], "--version")) {
			if (con_init() && reportConIOError()) return 1; // Init fail
			(*con_set_title)("KMatrix -- Matrix for KolibriOS");
			puts("KMatrix 1.1\n");
			puts("Copyright (C) 2021 Vitaliy Krylov\n");
			puts("License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.\n");
			puts("This is free software: you are free to change and redistribute it.\n");
			puts("This program comes with ABSOLUTELY NO WARRANTY.\n\n");
			puts("It is in no way affiliated in any way with the movie \"The Matrix\", \"Warner Bros\" nor any of its affiliates in any way.\n\n");
			puts("Written by Vitaliy Krylov.\n");
			(*con_exit)(0);
			return 0;
		} else if (!strcmp(argv[i], "--foreground") || !strcmp(argv[i], "-F")) {
			if (i + 1 < argc) {
				fgcolor_rgboverriden = false;
				if (!strcmp(argv[i+1], "black")) {
					fgcolor = 30;
				} else if (!strcmp(argv[i+1], "red")) {
					fgcolor = 31;
				} else if (!strcmp(argv[i+1], "green")) {
					fgcolor = 32;
				} else if (!strcmp(argv[i+1], "brown")) {
					fgcolor = 33;
				} else if (!strcmp(argv[i+1], "blue")) {
					fgcolor = 34;
				} else if (!strcmp(argv[i+1], "purple")) {
					fgcolor = 35;
				} else if (!strcmp(argv[i+1], "turqoise")) {
					fgcolor = 36;
				} else if (!strcmp(argv[i+1], "white")) {
					fgcolor = 37;
				} else {
					int rValue = -1, gValue = -1, bValue = -1;
					if (sscanf(argv[i+1], "%d,%d,%d", &rValue, &gValue, &bValue) != 3 ||
							rValue < 0 || rValue > 255 ||
							gValue < 0 || gValue > 255 ||
							bValue < 0 || bValue > 255) {
						if (con_init() && reportConIOError()) return 1; // Init fail
						(*con_set_title)("KMatrix -- Matrix for KolibriOS");
						printf("%s: invalid color: '%s'\n", argv[0], argv[i+1]);
						(*con_exit)(0);
						return 0;
					}
					fgcolor_rgboverriden = true;
					fgcolorrgb = rValue | (gValue << 8) | (bValue << 16);
				}
				i += 1;
			} else {
				if (con_init() && reportConIOError()) return 1; // Init fail
				(*con_set_title)("KMatrix -- Matrix for KolibriOS");
				printf("%s: option '%s' requires an argument\n", argv[0], argv[i]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				(*con_exit)(0);
				return 0;
			}
		} else if (!strcmp(argv[i], "--background") || !strcmp(argv[i], "-B")) {
			if (i + 1 < argc) {
				bgcolor_rgboverriden = false;
				if (!strcmp(argv[i+1], "black")) {
					bgcolor = 40;
				} else if (!strcmp(argv[i+1], "red")) {
					bgcolor = 41;
				} else if (!strcmp(argv[i+1], "green")) {
					bgcolor = 42;
				} else if (!strcmp(argv[i+1], "brown")) {
					bgcolor = 43;
				} else if (!strcmp(argv[i+1], "blue")) {
					bgcolor = 44;
				} else if (!strcmp(argv[i+1], "purple")) {
					bgcolor = 45;
				} else if (!strcmp(argv[i+1], "turqoise")) {
					bgcolor = 46;
				} else if (!strcmp(argv[i+1], "white")) {
					bgcolor = 47;
				} else {
					int rValue = -1, gValue = -1, bValue = -1;
					if (sscanf(argv[i+1], "%d,%d,%d", &rValue, &gValue, &bValue) != 3 ||
							rValue < 0 || rValue > 255 ||
							gValue < 0 || gValue > 255 ||
							bValue < 0 || bValue > 255) {
						if (con_init() && reportConIOError()) return 1; // Init fail
						(*con_set_title)("KMatrix -- Matrix for KolibriOS");
						printf("%s: invalid color: '%s'\n", argv[0], argv[i+1]);
						(*con_exit)(0);
						return 0;
					}
					bgcolor_rgboverriden = true;
					bgcolorrgb = rValue | (gValue << 8) | (bValue << 16);
				}
				i += 1;
			} else {
				if (con_init() && reportConIOError()) return 1; // Init fail
				(*con_set_title)("KMatrix -- Matrix for KolibriOS");
				printf("%s: option '%s' requires an argument\n", argv[0], argv[i]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				(*con_exit)(0);
				return 0;
			}
		} else if (!strcmp(argv[i], "--highlight") || !strcmp(argv[i], "-H")) {
			if (i + 1 < argc) {
				hlcolor_rgboverriden = false;
				if (!strcmp(argv[i+1], "black")) {
					hlcolor = 30;
				} else if (!strcmp(argv[i+1], "red")) {
					hlcolor = 31;
				} else if (!strcmp(argv[i+1], "green")) {
					hlcolor = 32;
				} else if (!strcmp(argv[i+1], "brown")) {
					hlcolor = 33;
				} else if (!strcmp(argv[i+1], "blue")) {
					hlcolor = 34;
				} else if (!strcmp(argv[i+1], "purple")) {
					hlcolor = 35;
				} else if (!strcmp(argv[i+1], "turqoise")) {
					hlcolor = 36;
				} else if (!strcmp(argv[i+1], "white")) {
					hlcolor = 37;
				} else {
					int rValue = -1, gValue = -1, bValue = -1;
					if (sscanf(argv[i+1], "%d,%d,%d", &rValue, &gValue, &bValue) != 3 ||
							rValue < 0 || rValue > 255 ||
							gValue < 0 || gValue > 255 ||
							bValue < 0 || bValue > 255) {
						if (con_init() && reportConIOError()) return 1; // Init fail
						(*con_set_title)("KMatrix -- Matrix for KolibriOS");
						printf("%s: invalid color: '%s'\n", argv[0], argv[i+1]);
						(*con_exit)(0);
						return 0;
					}
					hlcolor_rgboverriden = true;
					hlcolorrgb = rValue | (gValue << 8) | (bValue << 16);
				}
				i += 1;
			} else {
				if (con_init() && reportConIOError()) return 1; // Init fail
				(*con_set_title)("KMatrix -- Matrix for KolibriOS");
				printf("%s: option '%s' requires an argument\n", argv[0], argv[i]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				(*con_exit)(0);
				return 0;
			}
		} else if (!strcmp(argv[i], "--speed") || !strcmp(argv[i], "-S")) {
			if (i + 1 < argc) {
				int speedValue = 0;
				if (!sscanf(argv[i+1], "%d", &speedValue) || speedValue <= 0) {
					if (con_init() && reportConIOError()) return 1; // Init fail
					(*con_set_title)("KMatrix -- Matrix for KolibriOS");
					printf("%s: invalid speed: '%s'\n", argv[0], argv[i+1]);
					(*con_exit)(0);
					return 0;
				}
				matrixDelay = 100 / speedValue;
				i += 1;
			} else {
				if (con_init() && reportConIOError()) return 1; // Init fail
				(*con_set_title)("KMatrix -- Matrix for KolibriOS");
				printf("%s: option '%s' requires an argument\n", argv[0], argv[i]);
				printf("Try '%s --help' for more information.\n", argv[0]);
				(*con_exit)(0);
				return 0;
			}
		} else if (!strcmp(argv[i], "--shell")) {
			runInShell = true;
		} else {
			if (con_init() && reportConIOError()) return 1; // Init fail
			(*con_set_title)("KMatrix -- Matrix for KolibriOS");
			printf("%s: unrecognized option '%s'\n", argv[0], argv[i]);
			printf("Try '%s --help' for more information.\n", argv[0]);
			(*con_exit)(0);
			return 0;
		}
	}

	if (runInShell && (fgcolor_rgboverriden || bgcolor_rgboverriden || hlcolor_rgboverriden)) {
		if (con_init() && reportConIOError()) return 1; // Init fail
		(*con_set_title)("KMatrix -- Matrix for KolibriOS");
		printf("%s: it's not allowed to set RGB-color in console mode\n", argv[0]);
		printf("Try '%s --help' for more information.\n", argv[0]);
		(*con_exit)(0);
		return 0;
	}

	int bakedTable_size = '~'-'!';
	char bakedTable[bakedTable_size];
	for (char c = '!'; c < '\\'; ++c) bakedTable[c-'!'] = c;
	for (char c = '\\'+1; c <= '~'; ++c) bakedTable[c-'!'-1] = c;

	rand_r_seed = time(NULL);

	// I divide it in two parts because we don't want some additional operations in while(true)   (checkings if it's running in shell)
	if (runInShell) {
		if (con_init() && reportConIOError()) return 1; // Init fail
		(*con_set_title)("KMatrix -- Matrix for KolibriOS");
		int initCurHeight = (*con_get_cursor_height)();
		(*con_set_cursor_height)(0);
		printf("\033[%d;%dm", fgcolor, bgcolor);
		for (int i = 0; i < CONSOLE_WIDTH*CONSOLE_HEIGHT; ++i) putc(' ');
		bool lastWasHighlighted = false;

		Queue* q[CONSOLE_WIDTH/2];
		for (int x = 0; x < CONSOLE_WIDTH/2; ++x) q[x] = Q_alloc(sizeof(MatrixRow*), 0);
		MatrixRow* mr;
		while (true) {
			for (int x = 0; x < CONSOLE_WIDTH/2; ++x) {
				if (Q_size(q[x])) {
					Q_at(q[x], 0, &mr);
					if (mr->pos + 1 == CONSOLE_HEIGHT) {
						Q_pop(q[x], 0);
						free(mr->sequence);
 						free(mr);
					}
				}
				for (int i = 0; i < Q_size(q[x]); ++i) {
					Q_at(q[x], i, &mr);
					int idx = mr->pos + mr->length - 1;
					if (idx < CONSOLE_HEIGHT) {
						(*con_set_cursor_pos)(x*2, idx);
						char tc = (mr->sequence[0] == '\0') ? ' ' : mr->sequence[idx];
						if (!lastWasHighlighted) putc(tc);
						else { printf("\033[%dm%c", fgcolor, tc); lastWasHighlighted = 0; }
					}
					++(mr->pos);
					idx = mr->pos + mr->length - 1;
					if (idx < CONSOLE_HEIGHT) {
						(*con_set_cursor_pos)(x*2, idx);
						char tc = (mr->sequence[0] == '\0') ? ' ' : mr->sequence[idx];
						if (lastWasHighlighted) putc(tc);
						else { printf("\033[%dm%c", hlcolor, tc); lastWasHighlighted = 1; }
					}
				}

				bool needToAdd = !Q_size(q[x]);
				if (!needToAdd) {
					Q_at(q[x], Q_size(q[x]) - 1, &mr);
					needToAdd = (mr->pos > 0);
				}
				if (needToAdd) {
					bool mrseq0 = 0;
					if (Q_size(q[x])) mrseq0 = (mr->sequence[0] != '\0');
					mr = (MatrixRow*)malloc(sizeof(MatrixRow));
					if (!mr) {
						_ksys_debug_puts("[KMatrix] MatrixRow allocation error.\n");
						continue;
					}
					mr->length = (rand_r() % CONSOLE_HEIGHT) + 1;
					mr->pos = 1 - mr->length;
					if (!mrseq0 && (rand_r() % 3) == 2) { // 33% chance of being row with symbols
						mr->sequence = (char*)malloc(CONSOLE_HEIGHT);
						if (!mr->sequence) {
							_ksys_debug_puts("[KMatrix] MatrixRow->sequence allocation error.\n");
							free(mr);
							continue;
						}
						for (int i = 0; i < CONSOLE_HEIGHT; ++i)
							mr->sequence[i] = bakedTable[rand_r() % bakedTable_size];
					} else {
						mr->sequence = (char*)malloc(1);
						if (!mr->sequence) {
							_ksys_debug_puts("[KMatrix] MatrixRow->sequence allocation error.\n");
							free(mr);
							continue;
						}
 						mr->sequence[0] = '\0';
					}
					Q_push(q[x], &mr);
					(*con_set_cursor_pos)(x*2, 0);
					char tc = (mr->sequence[0] == '\0') ? ' ' : mr->sequence[0];
					if (lastWasHighlighted) putc(tc);
					else { printf("\033[%dm%c", hlcolor, tc); lastWasHighlighted = 1; }
				}
			}

			_ksys_delay(matrixDelay);
		}

		//I know it will not be executed.. but it better be executed when something sends STOP-like call when there will be better console.obj and others
		printf("\033[0m"); // Turning BG and FG to normal
		(*con_set_cursor_height)(initCurHeight);
		(*con_exit)(0);
	} else {
		if (!fgcolor_rgboverriden) {
			switch (fgcolor) {
				case 30:
					fgcolorrgb = 0x000000;
					break;
				case 31:
					fgcolorrgb = 0x800000;
					break;
				case 32:
					fgcolorrgb = 0x008000;
					break;
				case 33:
					fgcolorrgb = 0x808000;
					break;
				case 34:
					fgcolorrgb = 0x000080;
					break;
				case 35:
					fgcolorrgb = 0x800080;
					break;
				case 36:
					fgcolorrgb = 0x008080;
					break;
				case 37:
					fgcolorrgb = 0xc0c0c0;
					break;
			}
		}
		if (!bgcolor_rgboverriden) {
			switch (bgcolor) {
				case 40:
					bgcolorrgb = 0x000000;
					break;
				case 41:
					bgcolorrgb = 0x800000;
					break;
				case 42:
					bgcolorrgb = 0x008000;
					break;
				case 43:
					bgcolorrgb = 0x808000;
					break;
				case 44:
					bgcolorrgb = 0x000080;
					break;
				case 45:
					bgcolorrgb = 0x800080;
					break;
				case 46:
					bgcolorrgb = 0x008080;
					break;
				case 47:
					bgcolorrgb = 0xc0c0c0;
					break;
			}
		}
		if (!hlcolor_rgboverriden) {
			switch (hlcolor) {
				case 30:
					hlcolorrgb = 0x000000;
					break;
				case 31:
					hlcolorrgb = 0x800000;
					break;
				case 32:
					hlcolorrgb = 0x008000;
					break;
				case 33:
					hlcolorrgb = 0x808000;
					break;
				case 34:
					hlcolorrgb = 0x000080;
					break;
				case 35:
					hlcolorrgb = 0x800080;
					break;
				case 36:
					hlcolorrgb = 0x008080;
					break;
				case 37:
					hlcolorrgb = 0xc0c0c0;
					break;
			}
		}

		ksys_pos_t screenSize = _ksys_screen_size(); // WARNING: If user changes screen resolution while this program is running, here probably could occur bug
		screenSize.x += 1; screenSize.y += 1; // Yep, that's really needed :/

		int console_width = screenSize.x / SYSFN04_FONT_W;
		int console_height = screenSize.y / SYSFN04_FONT_H;

		char ss[2];
		ss[1] = '\0';

		Queue* q[console_width/2];
		for (int x = 0; x < console_width/2; ++x) q[x] = Q_alloc(sizeof(MatrixRow*), 0);
		MatrixRow* mr;

		int event;
		while (true) {
			event = _ksys_check_event();
			switch (event) {
				case KSYS_EVENT_NONE:
					break;
				case KSYS_EVENT_REDRAW: {
					_ksys_start_draw();
					_ksys_create_window(0, 0, screenSize.x, screenSize.y, "KMatrix -- Matrix for KolibriOS", bgcolorrgb, 0x31);
					_ksys_draw_bar(0, 0, screenSize.x, screenSize.y, bgcolorrgb);

					for (int x = 0; x < console_width/2; ++x) {
						for (int i = 0; i < Q_size(q[x]); ++i) {
							Q_at(q[x], i, &mr);
							if (mr->sequence[0] != '\0') {
								int y = mr->pos;
								if (y < 0) y = 0;
								for (; y < mr->pos + mr->length - 1 && y < console_height; ++y) {
									ss[0] = mr->sequence[y];
									_ksys_draw_text(ss, x*2*SYSFN04_FONT_W, y*SYSFN04_FONT_H, 0, 0x90000000 | fgcolorrgb);
								}
								if (y == mr->pos + mr->length - 1 && y < console_height) {
									ss[0] = mr->sequence[y];
									_ksys_draw_text(ss, x*2*SYSFN04_FONT_W, y*SYSFN04_FONT_H, 0, 0x90000000 | hlcolorrgb);
								}
							}
						}
					}

					_ksys_end_draw();
					break;
				}
				default:
					break;
			}

			for (int x = 0; x < console_width/2; ++x) {
				if (Q_size(q[x])) {
					Q_at(q[x], 0, &mr);
					if (mr->pos + 1 == console_height) {
						Q_pop(q[x], 0);
						free(mr->sequence);
						free(mr);
					}
				}
				for (int i = 0; i < Q_size(q[x]); ++i) {
					Q_at(q[x], i, &mr);
					int idx = mr->pos + mr->length - 1;
					if (idx < console_height) {
						_ksys_draw_bar(x*2*SYSFN04_FONT_W, idx*SYSFN04_FONT_H, SYSFN04_FONT_W, SYSFN04_FONT_H, bgcolorrgb);
						if (mr->sequence[0] != '\0') {
							ss[0] =  mr->sequence[idx];
							_ksys_draw_text(ss, x*2*SYSFN04_FONT_W, idx*SYSFN04_FONT_H, 0, 0x90000000 | fgcolorrgb);
						}
 					}
					++(mr->pos);
					idx = mr->pos + mr->length - 1;
					if (idx < console_height) {
						_ksys_draw_bar(x*2*SYSFN04_FONT_W, idx*SYSFN04_FONT_H, SYSFN04_FONT_W, SYSFN04_FONT_H, bgcolorrgb);
						if (mr->sequence[0] != '\0') {
							ss[0] =  mr->sequence[idx];
							_ksys_draw_text(ss, x*2*SYSFN04_FONT_W, idx*SYSFN04_FONT_H, 0, 0x90000000 | hlcolorrgb);
						}
					}
				}

				bool needToAdd = !Q_size(q[x]);
				if (!needToAdd) {
					Q_at(q[x], Q_size(q[x]) - 1, &mr);
					needToAdd = (mr->pos > 0);
				}
				if (needToAdd) {
					bool mrseq0 = 0;
					if (Q_size(q[x])) mrseq0 = (mr->sequence[0] != '\0');
					mr = (MatrixRow*)malloc(sizeof(MatrixRow));
					if (!mr) {
						_ksys_debug_puts("[KMatrix] MatrixRow allocation error.\n");
						continue;
					}
					mr->length = (rand_r() % console_height) + 1;
					mr->pos = 1 - mr->length;
					if (!mrseq0 && (rand_r() % 3) == 2) { // 33% chance of being row with symbols
						mr->sequence = (char*)malloc(console_height);
						if (!mr->sequence) {
							_ksys_debug_puts("[KMatrix] MatrixRow->sequence allocation error.\n");
							free(mr);
							continue;
						}
						for (int i = 0; i < console_height; ++i)
							mr->sequence[i] = bakedTable[rand_r() % bakedTable_size];
					} else {
						mr->sequence = (char*)malloc(1);
						if (!mr->sequence) {
							_ksys_debug_puts("[KMatrix] MatrixRow->sequence allocation error.\n");
							free(mr);
							continue;
						}
						mr->sequence[0] = '\0';
					}
					Q_push(q[x], &mr);
					_ksys_draw_bar(x*2*SYSFN04_FONT_W, 0, SYSFN04_FONT_W, SYSFN04_FONT_H, bgcolorrgb);
					if (mr->sequence[0] != '\0') {
						ss[0] =  mr->sequence[0];
 						_ksys_draw_text(ss, x*2*SYSFN04_FONT_W, 0, 0, 0x90000000 | hlcolorrgb);
					}
				}
			}

			_ksys_delay(matrixDelay);
		}
	}
	return 0;
}
